from mtmaod.products.modis_grid_rslab import MODISGrid

if __name__ == "__main__":
    path = r"C:\Users\imutu\Desktop\MOD021KM_L.1000.2021001040500.H26V05.000000.h5"
    with MODISGrid.open(path) as ds:
        print(MODISGrid.list_datasets(ds))
        d = MODISGrid.read(ds, "/AngleData/SolarAzimuthAngle")
        dp = MODISGrid.read(ds, "/AngleData/SolarAzimuthAngle").dp
        data = MODISGrid.read(ds, "/AngleData/SolarAzimuthAngle")[:]



import numpy as np
from sklearn.metrics import r2_score, mean_squared_error


def root_mean_square_error(label: list, predict: list) -> float:
    label = np.array(label)
    predict = np.array(predict)
    return np.sqrt(np.mean(np.square(predict - label)))


def Pearson(label: list, predict: list) -> float:
    label = np.array(label)
    predict = np.array(predict)
    return np.corrcoef(label, predict)[0, 1]


def envelope_of_expected_error(label: list, predict: list, relative_error=0.15, absolute_error=0.05) -> tuple:
    """
    预期误差: Expected Error
    预期误差包络线: Envelope of Expected Error

    Args:
        label (list): 真值列表
        predict (list): 非真值列表

    Returns:
        tuple: above_percent, below_percent, within_percent. EE包络线上百分比, EE包络线下百分比, EE包络线内百分比
    """
    above_count = 0
    below_count = 0
    within_count = 0
    label = np.array(label)
    predict = np.array(predict)
    # 判断输入是否合法
    if len(label) != len(predict) or len(label) == 0:
        return np.nan, np.nan, np.nan
    # 计算EE上下包络线
    label_above_envelope = label + (relative_error * label + absolute_error)
    label_below_envelope = label - (relative_error * label + absolute_error)
    # 获取EE包络线上下数组
    flag_above_array = predict > label_above_envelope
    flag_below_array = predict < label_below_envelope
    flag_within_array = np.logical_and(predict <= label_above_envelope, predict >= label_below_envelope)
    # 获取EE包络线上下的数量
    above_count = np.sum(flag_above_array)
    below_count = np.sum(flag_below_array)
    within_count = np.sum(flag_within_array)
    total_count = above_count + below_count + within_count
    # 计算EE包络线上下的百分比
    above_percent = above_count / total_count * 100
    below_percent = below_count / total_count * 100
    within_percent = within_count / total_count * 100
    return above_percent, below_percent, within_percent


def median_bias(label: list, predict: list) -> float:
    """
    中值偏差median bias

    Args:
        label (list): 真值列表
        predict (list): 非真值列表

    Returns:
        float: 结果
    """
    label = np.array(label)
    predict = np.array(predict)
    return np.median(predict) - np.median(label)


def mean_absolute_error(label: list, predict: list) -> float:
    """
    平均绝对误差mean absolute error

    Args:
        label (list): 真值列表
        predict (list): 非真值列表

    Returns:
        float: 结果
    """
    label = np.array(label)
    predict = np.array(predict)
    return np.mean(np.abs(predict - label))

if __name__ == '__main__':
    print(root_mean_square_error([-2, -1, 0, 1, 2], [4, 1, 3, 2, 0]))
    print(Pearson([-2, -1, 0, 1, 2], [4, 1, 3, 2, 0]))
    print(envelope_of_expected_error([-2, -1, 0, 1, 2], [4, 1, 3, 2, 0]))
    print(median_bias([-2, -1, 0, 1, 2], [4, 1, 3, 2, 0]))
    print(Pearson([1], [2]))


from mtmaod.products.mod04 import MOD04
import psutil, os
import numpy as np

path = r"E:\software\project\20230117185746\download\MOD04_L2.A2023263.0410.061.2023263132925\MOD04_L2.A2023263.0410.061.2023263132925.hdf"

ds = MOD04.open(path)
band_names = MOD04.list_datasets(ds)
print(band_names)

band_aod = "Optical_Depth_Land_And_Ocean"
data_aod = MOD04.read(ds, band_aod)[:]
print(data_aod.dtype, data_aod.shape)

data_lat = MOD04.read(ds, MOD04.Band_Latitude)[:]
print(data_lat.dtype, data_lat.shape)

data_lon = MOD04.read(ds, MOD04.Band_Longitude)[:]
print(data_lon)
print(data_lon.dtype, data_lon.shape)

print(u'当前进程的内存使用：%.4f GB' % (psutil.Process(os.getpid()).memory_info().rss / 1024 / 1024 / 1024))


from mtmaod.products.aeronet import LEVReader
import pandas as pd

if __name__ == "__main__":
    # 读取数据
    lev_file_path = r"E:\data-dbl\AOD_retrieval\AERONET\AERONET_lev15\19930101_20210102_Beijing_RADI.lev15"
    reader = LEVReader(lev_file_path, header=6)  # header为LEV文件的表头行索引, 默认文件第6行(从0开始)为表头行

    # Note: LEVReader实例的df变量即为读取的数据, LEVReader实例的方法若不指定dataframe参数, 则默认使用读取的df变量作为数据源
    # 筛选高质量数据
    df = reader.filter_high_quality_aod_rows()

    # 插值550nm波段的AOD, 当使用"scipy_curvefit" 或 "cubic+scipy"方法时, 调用多核CPU, 需要在"if __name__ == '__main__'"环境中使用
    df["AOD_550nm"] = reader.interp_aod_xxxnm(df, method="numpydeg2_polyfit", wavelength=550)

    # 对比不同AOD插值方法在500nm波段的表现
    reader.compare(dataframe=reader.df, filter_high_quality=True, wavelength=500)

    # 更新reader实例的df变量
    reader.df = df

    # 只保留AOD列
    df_aod = reader._only_keep_aod_column(df)

    # 插值到指定时间，构造时间戳数据框，注意时间戳的单位为秒
    # LEVReader的时区为UTC，因此依照时间插值时要注意产品的时区也应当是UTC时区的时间戳
    timestamps = pd.DataFrame.from_dict({"timestamp": [i.timestamp() + 900 for i in df.index]})
    # df_interp = reader.interp_aod_xxxtime(df["AOD_550nm"], objtime=timestamps, method="average")
    timestamps[["AOD_550nm", "counts"]] = reader.interp_aod_xxxtime(
        df["AOD_550nm"], objtime=timestamps, method="average"
    )
    print(timestamps)
    pass



from rasterio import Affine, CRS
import rasterio as rio
import numpy as np
from mtmaod.products.modis_grid_rslab import MODISGrid


def mod02grid_to_geotiff(src_file, dst_file, data, nodata=0):
    ds = MODISGrid.open(src_file)
    crs = CRS.from_string(ds.ProjectionStr)
    crs_transform = Affine.from_gdal(*map(float, ds.ProjectionPara.split(",")))
    profile = {}
    np.nan_to_num(data, copy=False, nan=nodata)
    # 更新TIFF参数
    profile.update(
        {
            "dtype": "float32",
            "width": data.shape[1],
            "height": data.shape[2],
            "count": data.shape[0],
            "compress": "lzw",
            "nodata": nodata,
            "crs": crs,
            "transform": crs_transform,
        }
    )
    with rio.open(dst_file, "w", **profile) as ds:
        ds.write(data)



from pyhdf.SD import SD, SDC, SDS


class PyHDF:
    DATATYPES = {
        4: "char",
        3: "uchar",
        20: "int8",
        21: "uint8",
        22: "int16",
        23: "uint16",
        24: "int32",
        25: "uint32",
        5: "float32",
        6: "float64",
    }
    OPENMODES = {"r": SDC.READ, "w": SDC.WRITE}
    @staticmethod
    def open(file_path:str, mode='r', *args, **kwargs):
        return SD(file_path, mode=PyHDF.OPENMODES[mode], *args, **kwargs)

    @staticmethod
    def read(fp:SD, dataset_name:str):
        return fp.select(dataset_name)
    
    @staticmethod
    def list_datasets(fp:SD, full=False):
        datasets_name = list(fp.datasets().keys())
        if not full:
            return datasets_name
        else:
            return {name: PyHDF.get_dataset_info_from_fp(fp, name) for name in datasets_name}

    @staticmethod
    def get_dataset_info_from_dp(dp:SDS):
        # https://hdfeos.github.io/pyhdf/modules/SD.html
        attrs = dp.attributes()
        _info_list = dp.info()
        _info_dict = {
            "dataset_name": _info_list[0], 
            "dataset_rank": _info_list[1], 
            "dataset_dims": _info_list[2], 
            "dataset_type": PyHDF.DATATYPES[_info_list[3]]
        }
        _info_dict.update(attrs)
        return _info_dict
     
    @staticmethod
    def get_dataset_info_from_fp(fp:SD, dataset_name:str):
        # https://hdfeos.github.io/pyhdf/modules/SD.html
        return PyHDF.get_dataset_info_from_dp(fp.select(dataset_name))
    



import os, glob
from mtmtool.time import auto_parse_time_with_datefmt
import pandas as pd
from datetime import timezone
import re



class Extractor:
    @staticmethod
    def file_datetime(datefmt:str, *args, **kwargs):
        """按照规则, 获取文件的时间, 并设置时区为UTC

        Parameters
        ----------
        datefmt : str, optional
            文件名中时间的格式, %Y为年, %m为月, %d为月日, %j为年日, %H为时, %M为分, %S为秒, 如果一个文件有多个时间匹配该规则, 取第一个时间为文件时间, by default "A%Y%j.%H%M" (MODIS系列文件的时间格式)
        """
        return lambda x: auto_parse_time_with_datefmt(x, datefmt=datefmt)[0].replace(tzinfo=timezone.utc)

    @staticmethod
    def file_hv(*args, **kwargs):
        """获取文件的HV分区, 额外支持hxxvxx

        """
        pattern = re.compile(r"[.](h(?:(?:\d{2})|(?:xx))v(?:(?:\d{2})|(?:xx)))[.]")
        return lambda x: re.findall(pattern, x.lower())[0]

        
def find_paths_with_extractor(root_dir:str, columns:list, indexes:dict, ext:str=".hdf")->pd.DataFrame:
    """寻找同一时刻的文件, 并返回一个DataFrame, 列名即为columns参数中的元素, 可以按需添加时间列

    Parameters
    ----------
    root_dir : str
        根目录, 该目录下的所有子文件夹都会被遍历
    columns : list
        查找的文件类型关键字, 如"MOD02", "MOD03"等
    indexes : dict
        将依据规则将提取结果添加到DataFrame中作为索引, 并以此合并不同的列。
        key为列名, value为一个函数, 该函数接受一个文件路径, 返回一个值
        value参数可以是mtmtool.path.Extractor中的静态方法, 也可以是自定义的函数
    ext : str, optional
        文件的后缀名称列表, 注意必须带'.', by default ".hdf" (MODIS系列文件的后缀名)

    Returns
    -------
    pd.DataFrame
        一个DataFrame, 列名即为columns参数中的元素, 索引为文件的时间, 无效数据为NaN
    """
    # 检查输入参数
    if not os.path.isdir(root_dir):
        raise ValueError(f"root_dir: {root_dir} 不是一个有效的文件夹路径")
    if not isinstance(columns, list) or len(columns) == 0:
        raise ValueError(f"columns: {columns} 不是一个有效的列表")
    if not isinstance(indexes, dict):
        raise ValueError(f"indexes: {indexes} 不是一个有效的字典")
    # 寻找同一时刻的文件, 并返回一个DataFrame, 列名即为columns参数中的元素
    df_list = []
    for file_type in columns: 
        # 保存临时数据的字典
        temp_data_dict = {}
        # 遍历所有子文件夹, 寻找符合条件的文件
        path = os.path.join(root_dir, "**", f"*{file_type}*{ext}") # 通配符 "**" 表示匹配0个或多个子文件夹
        for idx, _path in enumerate(glob.glob(path, recursive=True)) :
            
            _temp_dict = {}
            for k, v in indexes.items():
                try:
                    _temp_dict[k] = v(_path)
                except:
                    pass
            _temp_dict[file_type] = _path
            temp_data_dict[idx] = _temp_dict
        # 将字典转换为DataFrame, 方便后续合并不同列的数据
        _df = pd.DataFrame.from_dict(temp_data_dict, orient="index")
        indexes_names = list(indexes.keys())
        for i in indexes_names:
            if i not in _df.columns:
                _df[i] = None
        if len(indexes_names) > 0:
            _df = _df.set_index(indexes_names)
        df_list.append(_df)
    df = pd.concat(df_list, axis=1,  join='outer') # 按照索引合并不同列的数据
    return df


import numpy as np
from mtmaod.utils.pyhdf import PyHDF, SDS
import pandas as pd

from .template import SatelliteProductReader, SatelliteProductData


class MOD02_Data(SatelliteProductData):
    def __init__(self, dp: SDS, isRaw: bool = False, *args, **kwargs) -> None:
        self.dp = dp
        self.isRaw = isRaw
        pass

    def infos(self):
        return PyHDF.get_dataset_info_from_dp(self.dp)

    def scale_and_offset(self, data: np.ndarray):
        infos: dict = self.infos()
        radiance_scales = MOD02_Data.value_set_decimal(infos.get("reflectance_scales", 1), decimal=None)
        radiance_offsets = MOD02_Data.value_set_decimal(infos.get("reflectance_offsets", 0), decimal=None)
        fill_value = infos.get("_FillValue")
        data = data.astype(np.float64)
        data[data == fill_value] = np.nan
        return radiance_scales * (data - radiance_offsets)

    def __getitem__(self, *item):
        data = self.dp.__getitem__(*item)
        return self.scale_and_offset(data) if not self.isRaw else data


class MOD02(SatelliteProductReader):
    Product_File_Time_Format = "[.]A%Y%j[.]%H%M[.]"

    @staticmethod
    def open(data_file, *args, **kwargs):
        return PyHDF.open(data_file, *args, **kwargs)

    @staticmethod
    def read(fp, dataset_name, *args, isRaw=False, **kwargs):
        dp = PyHDF.read(fp, dataset_name, *args, **kwargs)
        return MOD02_Data(dp, isRaw=isRaw)

    @staticmethod
    def list_datasets(fp, full: bool = False, *args, **kwargs):
        return PyHDF.list_datasets(fp, full=full, *args, **kwargs)

    @staticmethod
    def table_scales_and_offsets(fp, *args, **kwargs):
        bands = ["EV_1KM_RefSB", "EV_1KM_Emissive", "EV_250_Aggr1km_RefSB", "EV_500_Aggr1km_RefSB"]
        columns = [
            "band_names",
            "reflectance_scales",
            "reflectance_offsets",
            "radiance_scales",
            "radiance_offsets",
            "corrected_counts_scales",
            "corrected_counts_offsets",
        ]
        indexes = [
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "10",
            "11",
            "12",
            "13lo",
            "13hi",
            "14lo",
            "14hi",
            "15",
            "16",
            "17",
            "18",
            "19",
            "20",
            "21",
            "22",
            "23",
            "24",
            "25",
            "26",
            "27",
            "28",
            "29",
            "30",
            "31",
            "32",
            "33",
            "34",
            "35",
            "36",
        ]
        df_list = []
        for band in bands:
            info = MOD02.read(fp, band, *args, **kwargs).infos()
            info["band_names"] = info.get("band_names").split(",")
            _info = {k: info[k] for k in columns if k in info}
            df_list.append(pd.DataFrame(_info))
        return pd.concat(df_list, ignore_index=True).set_index("band_names").loc[indexes, :]



from  __future__ import division
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from sklearn.metrics import r2_score,mean_squared_error
matplotlib.use('Agg')
#matplotlib.use('TkAgg')
from scipy import stats


# ===========Calculate the point density==========
x = np.arange(0, 2, 0.01)
y = np.arange(0, 2, 0.01)
xy = np.vstack([x, y])
# xy = xy.T
print(xy.shape)
print(np.max(xy, axis=1))
z = stats.gaussian_kde(xy)(xy)
# # ===========Sort the points by density, so that the densest points are plotted last===========
# idx = z.argsort()
# print(len(idx))
# x, y, z = x[idx], y[idx], z[idx]

# np.savetxt('alldata.csv', label_test, delimiter=',')


label_test = x
## r2与 MSE/RMSE
r2 = r2_score(x,y)
mse = mean_squared_error(x,y)
rmse = np.sqrt(mean_squared_error(x,y))
count = np.size(y)

plt.scatter(x,y, marker = 'o', c = z, s = 20, cmap = 'jet', vmin = 0)
plt.colorbar()    #显示颜色条


index=np.where(x>0.5)[0]
EEenvelops=len(index)/len(x)
print(EEenvelops)
max_X_value= 1.75
# my_x_ticks = np.arange(0, max_X_value, 0.25/2)
my_x_ticks = np.arange(0,max_X_value, 0.2)
# my_y_ticks = np.arange(0, max_X_value, 0.25/2)
my_y_ticks = np.arange(0, max_X_value,0.2)
plt.xticks(my_x_ticks)
plt.yticks(my_y_ticks)
plt.xlim(0, max_X_value)
plt.ylim(0, max_X_value)
#
# index=[]
# for i in range(len(x)):
#     if (y[i]>(1.15 * x[i] + 0.05) or y[i] < (0.85 * x[i] - 0.05)) and y[i]>0.5 and y[i]<0.85: #超过该范围就记为异常点，可以通过exception.py
#         index.append(i)
# index=np.array(index)
# np.save('temp/index.npy',index)


#计算测试集的R2，mse、rmse
# r2 = r2_score(x,y)
# mse = mean_squared_error(x,y)
# rmse = mse **0.5
# count = np.size(y)


#plt.xlim(0, max_X_value)
#plt.ylim(0, max_X_value)

x_hat = np.arange(0,max_X_value,0.001)
plt.plot(x_hat,x_hat,color='black', linestyle='solid')
plt.plot(x_hat, 0.85 * x_hat - 0.05, color='red', linestyle='dashed')
plt.plot(x_hat, 1.15 * x_hat + 0.05, color='red', linestyle='dashed')

index = np.where(y>(1.15 * x + 0.05))[0]
above=len(index)/len(label_test)
plt.text(0.03,1.45,'{:.2f}% above EE'.format(above*100),color='black')

index = np.where(y < (0.85 * x - 0.05))[0]
below=len(index)/len(label_test)
plt.text(0.03,1.35,'{:.2f}% below EE'.format(below*100),color='black')

index = np.where(y<=(1.15 * x + 0.05))[0]#within范围内的点
y=y[index]
x=x[index]
index = np.where(y >= (0.85 * x - 0.05))[0]
withEE=len(index)/len(label_test)
print(y.shape)
print(x.shape)
print(np.max(y))
print(np.max(x))



# withEE=0.85
plt.text(0.03,1.55,'{:.2f}% within EE'.format(withEE*100),color='blue')
plt.text(0.03,1.65,'EE envelopes:±(0.05+15%)',color='black')
# plt.grid(True)
plt.xlabel("AERONET AOD",fontdict={'size': 12})
plt.ylabel("SPAODnet MODIS AOD",fontdict={'size': 12})
plt.text(0.03,1.25,'R2 = {:.2f}'.format(r2),color='blue')
plt.text(0.03,1.15,'RMSE = {:.2f}'.format(rmse),color='blue')
plt.text(0.03,1.05,'N = {}'.format(count),color='blue')
# plt.scatter(x,y,color = 'blue', s = 100, marker = '*',alpha = 0.65)
#plt.scatter(x,y,s = 20, marker = 'o', alpha = 0.6, c=z, cmap = 'Spectral_r')        #viridis是调色盘


plt.savefig('result.png',dpi=600)
plt.show()



from mtmaod.products.mod02 import MOD02, PyHDF
import psutil, os
import numpy as np

path = r"C:\Users\imutu\Desktop\MOD021KM.A2021001.0405.061.2021001130625.hdf"

ds = MOD02.open(path)
band_names = MOD02.list_datasets(ds)
print(band_names)

band_1KM_RefSB = "EV_1KM_RefSB"
data_1KM_RefSB = MOD02.read(ds, band_1KM_RefSB)
table = MOD02.table_scales_and_offsets(ds)
# band_aod = "Optical_Depth_Land_And_Ocean"
# data_aod = MOD04.read(ds, band_aod)[:]
# print(data_aod.dtype, data_aod.shape)

# data_lat = MOD04.read(ds, MOD04.Band_Latitude)[:]
# print(data_lat.dtype, data_lat.shape)

# data_lon = MOD04.read(ds, MOD04.Band_Longitude)[:]
# print(data_lon)
# print(data_lon.dtype, data_lon.shape)

# print(u'当前进程的内存使用：%.4f GB' % (psutil.Process(os.getpid()).memory_info().rss / 1024 / 1024 / 1024))

pass


import pandas as pd
import numpy as np
import warnings

warnings.filterwarnings("ignore", message=".*Covariance of the parameters could.*")

# 检查swifter模块是否开启
try:
    from swifter import set_defaults

    set_defaults(progress_bar=False)
    _swifter_enable = True
except ImportError:
    _swifter_enable = False
    pass


def group_arr_by_nan(arr):
    """将数组按nan分组, 返回字典, key为非nan的索引, value为对应的行索引"""
    index_dict = {}
    for idx, t in enumerate(arr):
        a = ",".join(map(str, np.where(~np.isnan(t))[0]))
        if a in index_dict:
            index_dict[a].append(idx)
        else:
            index_dict[a] = [idx]
    return index_dict


class LEVReader:
    """
    LEVReader实例的df变量即为读取的数据

    LEVReader实例的方法若不指定dataframe参数, 则默认使用实例的df变量作为数据源
    """

    def __init__(self, filepath, header=4) -> None:
        with open(filepath, "r", encoding="Windows 1252") as f:
            txt = f.read()
        txt = txt.replace("  ", " ").replace("  ", " ").replace("  ", " ")
        with open(filepath, "w", encoding="Windows 1252") as f:
            f.write(txt)
        self.df = pd.read_csv(filepath, header=header, sep=" ", encoding="Windows 1252")
        self.df = self.df.replace(-999.0, np.nan)
        self.df = self._convert_time_to_utc()

    def _get_dataframe(self, dataframe=None):
        if dataframe is None:
            df = self.df.copy()
        else:
            df = dataframe.copy()
        return df

    def filter_high_quality_aod_rows(self, dataframe=None):
        """筛选出质量较高的AOD数据行, 要求AOD_440nm, AOD_500nm, AOD_675nm, AOD_870nm均大于0

        Parameters
        ----------
        dataframe : DataFrame
            数据源

        Returns
        -------
        DataFrame
            筛选后的数据
        """
        df = self._get_dataframe(dataframe)
        df = df[df["440nm"] > 0.0]
        df = df[df["500nm"] > 0.0]
        df = df[df["675nm"] > 0.0]
        df = df[df["870nm"] > 0.0]
        return df.copy()

    def _convert_time_to_utc(self, dataframe=None):
        df = self._get_dataframe(dataframe)
        df.loc[:, "CustomDate"] = df.loc[:, "date"] + " " + df.loc[:, "time(UTC)"]
        df.loc[:, "CustomDate"] = pd.to_datetime(df.loc[:, "CustomDate"], format="%d/%m/%Y %H:%M:%S", utc=True)
        df.set_index("CustomDate", inplace=True)
        df = df.sort_index()
        return df

    def _only_keep_aod_column(self, dataframe=None):
        """只保留AOD列, 并将列名改为波长名"""
        df = self._get_dataframe(dataframe)
        # 删除非AOD以及无效的数据列
        columns = [title for title in df.columns if title.endswith("nm")]
        df = df.loc[:, columns]
        df = df.replace(-999.0, np.nan)
        df = df.dropna(axis=1, how="all")
        # 将数据列名改为波长名
        columns = df.columns
        columns = [i.replace("AOD_", "").replace("nm", "") for i in columns]
        df.columns = columns
        # 根据波长大小对列名进行排序
        try:
            columns = [(int(i), i) for i in columns]
        except Exception as e:
            print("Aeronet数据格式与代码不同,请修改!")
            raise e
        columns = sorted(columns, key=lambda x: x[0])
        _, columns = zip(*columns)
        df = df.loc[:, columns]
        return df.copy()

    def interp_aod_xxxnm(self, dataframe=None, method="numpydeg2_polyfit", wavelength: float = 550) -> np.ndarray:
        """插值AOD到xxxnm波段

        Parameters
        ----------
        dataframe : pandas.DataFrame, optional
            输入的原始数据, 默认值为 None
        method : str, optional
            插值方法, 默认值为 "numpydeg2_polyfit". 可选值有: "linear","nearest", "nearest-up", "zero_spline", "slinear_spline", "quadratic_spline", "cubic_spline", "numpydeg2_polyfit", "numpydeg3_polyfit", "scipy_curvefit", "Angstrom", "cubic+scipy"

        wavelength : float, optional
            要插值的目标波段, 默认值为 550

        Returns
        -------
        np.ndarray
            插值后的AOD值
        """
        dataframe = self._get_dataframe(dataframe)
        if method in [
            "linear",
            "nearest",
            "nearest-up",
            "zero_spline",
            "slinear_spline",
            "quadratic_spline",
            "cubic_spline",
        ]:
            from scipy.interpolate import interp1d

            # 采用scipy的插值方法
            df_aod = self._only_keep_aod_column(dataframe.copy())
            x = np.array(list(map(int, df_aod.columns)))
            y = df_aod.to_numpy()
            method_real = method.replace("_spline", "")  # 为了兼容scipy的插值参数格式, 去掉_spline后缀
            group_dict = group_arr_by_nan(y)
            series_list = []
            for key in group_dict:
                columns = list(map(int, key.split(",")))
                x_temp = x[columns]
                if (method_real == "cubic" and len(x_temp) < 4) or min(x_temp) > wavelength or max(x_temp) < wavelength:
                    yfit_temp = np.full(len(group_dict[key]), np.nan)
                    series_list.append(pd.Series(data=yfit_temp, index=group_dict[key]))
                    continue
                y_temp = y[group_dict[key], :][:, columns]
                f = interp1d(x=x_temp, y=y_temp, kind=method_real, bounds_error=False)
                yfit_temp = f(wavelength)
                series_list.append(pd.Series(data=yfit_temp, index=group_dict[key]))
            yfit = pd.concat(series_list).sort_index()
        elif method in ["scipy_curvefit"]:
            from scipy.optimize import curve_fit

            def quadratic_polynomial_formula(x, a0, a1, a2):
                """二次多项式公式, 用于scipy.optimize.curve_fit(), 用于拟合AOD数据"""
                return np.exp(a0 + a1 * np.log(x) + a2 * np.log(x) ** 2)

            def polyfit2(df, wavelength=550, x=None):
                """单个数据行的AOD插值函数, 调用quadratic_polynomial_formula"""
                # 提取要拟合的数据列
                y = df.to_numpy()
                params, _ = curve_fit(quadratic_polynomial_formula, x, y, check_finite=False, nan_policy="omit")
                # print(params)
                yfit = quadratic_polynomial_formula(wavelength, *params)
                return yfit

            # https://blog.csdn.net/weixin_42182090/article/details/131018277
            df_aod = self._only_keep_aod_column(dataframe.copy())
            x = np.array(list(map(int, df_aod.columns)))
            if _swifter_enable:
                yfit = df_aod.swifter.apply(polyfit2, axis=1, wavelength=wavelength, x=x)
            else:
                warnings.warn("no module 'swifter' for useing multi cpus.")
                yfit = df_aod.apply(polyfit2, axis=1, wavelength=wavelength, x=x)
        elif method in ["numpydeg2_polyfit"]:
            df_aod = self._only_keep_aod_column(dataframe.copy())
            aod_columns = np.array(list(map(int, df_aod.columns)))
            x = np.log(np.array(list(map(int, df_aod.columns))))
            y = np.log(np.where(df_aod.to_numpy() <= 0, np.nan, df_aod.to_numpy()))
            # 分组拟合,防止nan值影响拟合结果
            group_dict = group_arr_by_nan(y)
            series_list = []
            for key in group_dict:
                idx_columns = list(map(int, key.split(",")))
                num_columns = aod_columns[idx_columns]
                x_temp = x[idx_columns]
                if len(x_temp) < 3 or min(num_columns) > wavelength or max(num_columns) < wavelength:
                    yfit_temp = np.full(len(group_dict[key]), np.nan)
                    series_list.append(pd.Series(data=yfit_temp, index=group_dict[key]))
                    continue
                y_temp = y[group_dict[key], :][:, idx_columns]
                coef = np.polyfit(x_temp, y_temp.T, deg=2)
                yfit_temp = np.exp(np.polyval(coef, np.log(wavelength)))
                series_list.append(pd.Series(data=yfit_temp, index=group_dict[key]))
            yfit = pd.concat(series_list).sort_index()
        elif method in ["numpydeg3_polyfit"]:
            df_aod = self._only_keep_aod_column(dataframe.copy())
            aod_columns = np.array(list(map(int, df_aod.columns)))
            x = np.log(np.array(list(map(int, df_aod.columns))))
            y = np.log(np.where(df_aod.to_numpy() <= 0, np.nan, df_aod.to_numpy()))
            # 分组拟合,防止nan值影响拟合结果
            group_dict = group_arr_by_nan(y)
            series_list = []
            for key in group_dict:
                idx_columns = list(map(int, key.split(",")))
                num_columns = aod_columns[idx_columns]
                x_temp = x[idx_columns]
                if len(x_temp) < 4 or min(num_columns) > wavelength or max(num_columns) < wavelength:
                    yfit_temp = np.full(len(group_dict[key]), np.nan)
                    series_list.append(pd.Series(data=yfit_temp, index=group_dict[key]))
                    continue
                y_temp = y[group_dict[key], :][:, idx_columns]
                coef = np.polyfit(x_temp, y_temp.T, deg=3)
                yfit_temp = np.exp(np.polyval(coef, np.log(wavelength)))
                series_list.append(pd.Series(data=yfit_temp, index=group_dict[key]))
            yfit = pd.concat(series_list).sort_index()
        elif method in ["Angstrom"]:
            df_raw = dataframe.copy()
            df_raw["440-675_Angstrom_Exponent"] = df_raw["440-675_Angstrom_Exponent"].astype(float)
            yfit = df_raw.apply(
                lambda x: x["AOD_440nm"] * np.power((wavelength / 440), -x["440-675_Angstrom_Exponent"]), axis=1
            )
        elif method in ["cubic+scipy"]:
            yfit_scipy = self.interp_aod_xxxnm(dataframe=dataframe, method="scipy_curvefit", wavelength=wavelength)
            yfit_cubic = self.interp_aod_xxxnm(dataframe=dataframe, method="cubic_spline", wavelength=wavelength)
            yfit = (yfit_scipy + yfit_cubic) / 2
        else:
            raise ValueError("method参数错误")
        return np.array(yfit)

    def interp_aod_xxxtime(
        self,
        srcdata: pd.DataFrame,
        objtime: pd.DataFrame,
        method: str = "average",
        time_range: tuple = (-1800, 1800),
    ) -> pd.DataFrame:
        """插值AOD到指定时间

        Parameters
        ----------
        srcdata : pd.DataFrame
            原始的数据,一般为Aeronet站点数据, 默认值为 None
        objtime : pd.DataFrame
            目标时刻数据,一般为卫星产品的过境时刻,必须含有"timestamp"列,存储UTC时刻,或者数据的索引为UTC的pandas.datatime64类型
        method : str, optional
            插值的方法,有"average"和"linear"两种,average会取一段时间内数据平均,linear会进行相邻时刻数据的线性插值, 默认值为"average"
        time_range : tuple, optional
            时间范围,单位为秒,以目标时刻为基准(目标时刻为0)。默认值为(-1800, 1800),-1800表示目标时刻前1800秒,1800表示目标时刻后1800秒,将在这前后半小时内进行插值操作。

        Returns
        -------
        pd.DataFrame
            插值后的数据,包含有两列,第一列为插值后的AOD值,第二列为插值的计数,即插值用到的数据个数。数据的索引为数据的排序数目,即第几个数据。
        """
        # dataframe 必须 包含 timestamp 列
        # timestamps 必须 包含 timestamp 列
        srcdata = srcdata.copy()
        # 检查输入参数
        if isinstance(srcdata, pd.Series):
            srcdata = pd.DataFrame(srcdata)
        assert isinstance(srcdata, pd.DataFrame), "srcdata 必须是一个Pandas 中的 DataFrame 或其子类"
        assert isinstance(objtime, pd.DataFrame), "objtime 必须是一个Pandas 中的 DataFrame 或其子类"
        if (
            pd.api.types.is_datetime64_dtype(srcdata.index) or pd.api.types.is_datetime64tz_dtype(srcdata.index)
        ) and "timestamp" not in srcdata.columns:
            srcdata["timestamp"] = [i.timestamp() for i in srcdata.index]
        assert "timestamp" in srcdata.columns, "srcdata 必须 包含 timestamp 列"
        assert "timestamp" in objtime.columns, "objtime 必须 包含 timestamp 列"

        # # 第一步： 获取以目标时间为中心的时间段内的原始数据
        from shapely.geometry import box
        import geopandas as gpd

        # 构造时间box几何序列
        pre_time_offset, post_time_offset = time_range
        gs_obj = gpd.GeoSeries([box(i + pre_time_offset, -1, i + post_time_offset, 1) for i in objtime["timestamp"]])
        # 原始数据点转为point几何
        gs_src = gpd.GeoSeries(gpd.points_from_xy(srcdata["timestamp"], 0))
        # 获取查找表,此时的lookup_table是一个二维数组, 每一列表示一个目标和来源的匹配关系,第一行是目标的索引,第二行是来源的索引
        lookup_table = gs_src.sindex.query(gs_obj, predicate="intersects")

        # # 第二步： 根据匹配关系,插值目标时间点的AOD值
        df_src = srcdata.iloc[lookup_table[1]].copy()
        df_src["_idx"] = lookup_table[0]
        if method == "average":
            df_src.drop(columns=["timestamp"], inplace=True)
            result_temp = df_src.groupby("_idx").mean()
            counts = df_src.groupby("_idx").count()
            result_temp["counts"] = counts
            result_temp.index = np.array(objtime.index)[np.unique(lookup_table[0])]  # 将索引改为目标时间点的索引
        elif method == "linear":
            # 筛选能被插值的数据时间
            counts = df_src.groupby("_idx").count().iloc[:, 0]
            counts = counts[(counts > 1).values]
            objtime_interp = objtime.iloc[counts.index]
            # 插值
            result_temp = pd.DataFrame(index=objtime_interp.index)
            for column in df_src.columns:
                if column in ["timestamp", "_idx", "counts"]:
                    continue
                result_temp[column] = np.interp(objtime_interp["timestamp"], srcdata["timestamp"], srcdata[column])
            # 添加计数列,由于是线性插值,所以计数列的值都是2
            result_temp["counts"] = 2
        else:
            raise ValueError("method参数错误")
        # 为插值失败的数据添加计数列,计数列的值为0
        result = pd.DataFrame(index=objtime.index, columns=result_temp.columns)
        result["counts"] = 0
        result.update(result_temp)
        return result

    def compare(self, dataframe=None, filter_high_quality=True, wavelength=500, methods=None):
        """对比不同AOD插值方法在xxxnm波段的表现, 并输出均值、最大值、耗时, 默认使用500nm波段"""
        import time

        df = self._get_dataframe(dataframe)
        df_src = df.query(f"{wavelength}nm > 0.0")
        if filter_high_quality:
            df_src = df_src.query("440nm > 0.0")
            df_src = df_src.query("500nm > 0.0")
            df_src = df_src.query("675nm > 0.0")
            df_src = df_src.query("870nm > 0.0")
        print("数据筛选与否: filter_high_quality = " + str(filter_high_quality))
        df = pd.DataFrame(index=df_src.index)
        df["true"] = df_src[f"{wavelength}nm"]
        df_src = df_src.drop(columns=[f"{wavelength}nm"])

        if methods is None:
            methods = ["Angstrom", "quadratic_spline", "cubic_spline", "numpydeg2_polyfit", "numpydeg3_polyfit"]

        for method in methods:
            name = method.split("_")[0]
            t_start = time.time()
            df[name] = self.interp_aod_xxxnm(dataframe=df_src, method=method, wavelength=wavelength)
            t_spended = time.time() - t_start
            bias = np.abs(df[name] - df["true"])
            print("{:15s}: 均误差{:10.6f}, 最大误差{:10.6f}, 耗时{:6.3f}s".format(name, bias.mean(), bias.max(), t_spended))



import netCDF4 as nc


class NetCDF4:
    @staticmethod
    def open(file_path, mode="r", *args, **kwargs):
        return nc.Dataset(file_path, mode=mode)
    
    @staticmethod
    def read(fp, dataset_name, *args, **kwargs):
        return NetCDF4._jump(fp, dataset_name)

    @staticmethod
    def list_datasets(fp, full=False):
        datasets_name = list(NetCDF4._walk(fp))
        if not full:
            return datasets_name
        else:
            return {name: NetCDF4.get_dataset_info_from_fp(fp, name) for name in datasets_name}
        
    @staticmethod
    def _walk(fp, path=""):
        if not len(path) or path[-1] != "/":
            path += "/"
        current_variables = list(fp.variables.keys())
        for variable in current_variables:
            yield path + variable
        current_groups = list(fp.groups.keys())
        for group in current_groups:
            yield from NetCDF4._walk(fp.groups[group], path + group)

    @staticmethod
    def _jump(fp, path="/"):
        path_list = path.lstrip("/").split("/")
        if not len(path_list):
            return fp
        subnode_fp = fp.__getitem__(path_list[0])
        subnode_path = "/" + "/".join(path_list[1:])
        if len(path_list) == 1:
            return subnode_fp
        else:
            return NetCDF4._jump(subnode_fp, subnode_path)
        
    @staticmethod
    def get_dataset_info_from_dp(dp):
        info_dict = dp.__dict__
        info_dict.update({
            "dataset_name": dp.group().path + "/" + dp.name,
            "dataset_dims": dp.shape,
            "dataset_type": dp.datatype.name
        })
        return info_dict

    @staticmethod
    def get_dataset_info_from_fp(fp, dataset_name):
        return NetCDF4.get_dataset_info_from_dp(NetCDF4._jump(fp, dataset_name))


from osgeo import gdal


path = r"C:\Users\imutu\Desktop\MOD021KM.A2021001.0405.061.2021001130625.hdf"
datasets = gdal.Open(path)
SubDatasets = datasets.GetSubDatasets()
SubDatasetsNum =  len(datasets.GetSubDatasets())
#  获取hdf中的元数据
Metadata = datasets.GetMetadata()
#  获取元数据的个数
MetadataNum = len(Metadata)
#  输出各子数据集的信息
print("元数据一共有{0}个: ".format(MetadataNum))
for key,value in Metadata.items():
    print('{key}:{value}'.format(key = key, value = value))


import numpy as np


class SatelliteProductData:
    def __init__(self) -> None:
        pass

    def infos(self):
        pass

    @staticmethod
    def value_set_decimal(values, decimal=None):
        values = np.array(values)
        values_int = values.astype(int)
        values_decimal = np.round(values, decimal) if decimal is not None else values
        return values_int if values_int == values else values_decimal

    def __getitem__(self, *item):
        pass


class SatelliteProductReader:
    Product_File_Time_Format = None
    Band_Latitude = None
    Band_Longitude = None

    @staticmethod
    def open(data_file, *args, **kwargs):
        pass

    @staticmethod
    def read(fp, dataset_name, *args, **kwargs):
        pass

    @staticmethod
    def list_datasets(fp, full: bool = False, *args, **kwargs):
        pass



import numpy as np
from mtmaod.utils.netCDF4 import NetCDF4

from .template import SatelliteProductReader, SatelliteProductData


class MODISGrid_Data(SatelliteProductData):
    def __init__(self, dp, isRaw: bool = False, *args, **kwargs) -> None:
        self.dp = dp
        self.isRaw = isRaw
        pass

    def infos(self):
        return NetCDF4.get_dataset_info_from_dp(self.dp)

    def scale_and_offset(self, data: np.ndarray):
        infos: dict = self.infos()
        fill_value = infos.get("FillValue")
        if isinstance(fill_value, str):
            fill_value = float(fill_value)
        data = data.astype(np.float64).data
        data[data == fill_value] = np.nan
        return data

    def __getitem__(self, *item):
        data = self.dp.__getitem__(*item)
        return self.scale_and_offset(data) if not self.isRaw else data.data


class MODISGrid(SatelliteProductReader):
    Product_File_Time_Format = "[.]%Y%j%H%M%S[.]"  # MOD021KM_L.1000.2021001040500.H26V05.000000.h5

    @staticmethod
    def open(data_file, *args, **kwargs):
        return NetCDF4.open(data_file, *args, **kwargs)

    @staticmethod
    def read(fp, dataset_name, *args, isRaw=False, **kwargs):
        dp = NetCDF4.read(fp, dataset_name, *args, **kwargs)
        return MODISGrid_Data(dp, isRaw=isRaw)

    @staticmethod
    def list_datasets(fp, full: bool = False, *args, **kwargs):
        return NetCDF4.list_datasets(fp, full=full, *args, **kwargs)



from mtmaod.plot import density_chart
import numpy as np


if __name__ == "__main__":
    x = np.arange(0, 2, 0.01)
    y = np.arange(0, 2, 0.01)

    # Note: if use non-GUI backend, then save_path must be specified
    # from mtmaod import mpl
    # mpl.use("Agg")

    # save path
    density_chart(x=x, y=y, save_path="demo_scatter.png", type="scatter", bins=25, dpi=300)
    density_chart(x=x, y=y, save_path="demo_pcolormesh.png", type="grid", bins=10, dpi=300, style="single_bold")
    density_chart(x=x, y=y, save_path="demo_kernels.png", type="kernels", dpi=300)

    # show in GUI
    density_chart(x=x, y=y, type="kernels", dpi=300)



import matplotlib as mpl
from matplotlib import pyplot as plt
import numpy as np
import os

from .indicator import envelope_of_expected_error, root_mean_square_error, r2_score


font_config = {
    "Default": {},
    "Times New Roman": {
        "family": "serif",
        "serif": ["Times New Roman"],
        "weight": "normal",
    },
}

style_config = {
    "single_normal": {"figsize": (3.538, 3.538), "fontsize": 10.5, "ticksize": 9},
    "single_bold": {"figsize": (5, 4.5), "fontsize": 12, "ticksize": 10.5},
    # "double_normal": {"figsize": (6.7, 5)},
    # "double_bold": {"figsize": (7.48, 6)},
}


def density_chart(
    x,
    y,
    xlabel="AOD Measured",
    ylabel="AOD Retrieved",
    title=None,
    type="kernels",
    bins=25,
    xy_max_edge=2,
    save_path=None,
    dpi=300,
    figsize=(3, 3),
    style=None,
    fontsize=9,
    ticksize=8,
):
    """绘制密度图, 来比较标签和预测值的分布情况, 使用高斯核密度(kernels)来绘制密度图 或 使用直方图(scatter or grid)来统计单位面积内的点的数量

    Parameters
    ----------
    x : list
        label
    y : list
        predict
    xlabel : str, optional
        x轴标签, by default "AOD Measured"
    ylabel : str, optional
        y轴标签, by default "AOD Retrieved"
    title : str, optional
        标题, by default None
    type : str, {"kernels", "scatter", "grid"}, optional
        要绘制的图的类型, by default "kernels"
    bins : int, optional
        直方图统计时, 每1个单位有多少个箱子, 每个箱子单位刻度为(1/bins), by default 25
    xy_max_edge : int, optional
        最大的xy显示边界, by default 2
    save_path : _type_, optional
        保存路径, 如果为None, 则直接显示, by default None
    dpi : int, optional
        分辨率, 每英寸下多少个像元, by default 300
    figsize : tuple, optional
        画布大小, 单位是英寸, by default (3, 3)
    style : _type_, optional
        风格样式, by default None
    fontsize : int, optional
        标题等的字体大小, by default 9
    ticksize : int, optional
        tick等的字体大小, by default 8

    """
    # 设置风格
    style = "single_bold" if style is None else style
    if style in style_config:
        figsize = style_config[style].get("figsize", figsize)
        fontsize = style_config[style].get("fontsize", fontsize)
        ticksize = style_config[style].get("ticksize", ticksize)
        fontname = style_config[style].get("fontname", "Times New Roman")
    # 设置字体, 默认使用Times New Roman, 如果系统中没有该字体, 则使用Matplotlib默认字体
    matplotlib_fontnames = [i.name for i in mpl.font_manager.fontManager.ttflist]
    if (fontname not in ["Default"]) and (fontname not in matplotlib_fontnames):
        print(f"Warning: 系统中没有找到{fontname}字体, 请安装该字体, 本次将使用默认字体")
        fontname = "Default"
    with mpl.rc_context():
        # 设置字体
        mpl.rc("font", **font_config[fontname])
        mpl.rcParams["font.size"] = fontsize
        mpl.rcParams["mathtext.fontset"] = "stix"

        # 设置布局
        mpl.rcParams["savefig.bbox"] = "tight"
        mpl.rcParams["figure.constrained_layout.use"] = True

        # Set x axis
        mpl.rcParams["xtick.direction"] = "in"
        mpl.rcParams["xtick.major.size"] = 3
        mpl.rcParams["xtick.major.width"] = 0.5
        mpl.rcParams["xtick.minor.size"] = 1.5
        mpl.rcParams["xtick.minor.width"] = 0.5
        mpl.rcParams["xtick.minor.visible"] = True
        mpl.rcParams["xtick.top"] = True

        # Set y axis
        mpl.rcParams["ytick.direction"] = "in"
        mpl.rcParams["ytick.major.size"] = 3
        mpl.rcParams["ytick.major.width"] = 0.5
        mpl.rcParams["ytick.minor.size"] = 1.5
        mpl.rcParams["ytick.minor.width"] = 0.5
        mpl.rcParams["ytick.minor.visible"] = True
        mpl.rcParams["ytick.right"] = True

        # 设置分辨率
        mpl.rcParams["figure.dpi"] = dpi  # plt.show显示分辨率
        mpl.rcParams["savefig.dpi"] = dpi  # plt.savefig保存分辨率

        # 设置画布
        fig, axes = plt.subplots(1, 1, figsize=figsize)

        # 设置其它参数
        axes.set_title(title)  # 设置标题
        # 设置坐标轴参数
        axes.set_xlabel(xlabel, labelpad=2)
        axes.set_ylabel(ylabel, labelpad=2)
        axes.set_xlim(-0.05, xy_max_edge + 0.05)
        axes.set_ylim(-0.05, xy_max_edge + 0.05)
        axes.set_aspect("equal")
        axes.grid(linestyle=":", color="r", alpha=0.1)
        # 设置刻度, 保证刻度不会超过xy_max_edge, 且刻度间隔为0.5
        xyticks = [i / 2 for i in range(int(xy_max_edge // 0.5) + 1)]
        xyticks = [i for i in xyticks if i <= xy_max_edge]
        xytickslables = [f"{format(i, '.1f')}" for i in xyticks]
        axes.set_xticks(xyticks)
        axes.set_yticks(xyticks)
        axes.set_xticklabels(xytickslables)
        axes.set_yticklabels(xytickslables)
        # 设置刻度字体大小
        axes.tick_params(labelsize=ticksize)

        # 将数据转为numpy数组
        x, y = np.array(x).reshape(-1), np.array(y).reshape(-1)  # x means label, y means predict
        mask = (x <= xy_max_edge) & (y <= xy_max_edge)
        print(f"数据总数: {len(x)}, 有效数据总数: {mask.sum()}, 越界数据总数: {(~mask).sum()}")

        if type in ["kernels"]:
            # 计算数据的高斯核密度分布
            from scipy.stats import gaussian_kde

            xy = np.vstack([x, y])
            z = gaussian_kde(xy)(xy)
            # 计算scatter散点的大小
            point_size = dpi / 100
            # 生成散点图图层
            ax_data = axes.scatter(x, y, marker="o", c=z, s=point_size, cmap="jet", vmin=0)

        elif type in ["scatter", "grid"]:
            # 计算数据的直方图分布
            xy_max_edge = xy_max_edge
            nbins = xy_max_edge * bins
            x = x[mask]
            y = y[mask]
            H, xedges, yedges = np.histogram2d(x, y, bins=nbins, range=[[0, xy_max_edge], [0, xy_max_edge]])

            if type == "scatter":  # 散点图
                x_index = (x * bins).astype("int")
                y_index = (y * bins).astype("int")
                z = H[x_index, y_index]
                # 获取散点大小
                axis_height = axes.xaxis.clipbox.height  # 获取坐标轴高度
                axis_length = axes.get_xlim()[1] - axes.get_xlim()[0]  # 获取坐标轴长度
                scatter_point_size = int(axis_height / (bins * (axis_length)))
                s = (scatter_point_size**2) / 4
                # 画散点图
                sort_index = np.argsort(z)
                x = x[sort_index]
                y = y[sort_index]
                z = z[sort_index]
                # 生成散点图图层
                ax_data = axes.scatter(x, y, marker=".", c=z, s=s, label="AOD scatter", cmap="jet", linewidths=0)
            elif type == "grid":  # 方格图
                # H needs to be rotated and flipped
                H = np.rot90(H)
                H = np.flipud(H)
                # Mask zeros
                Hmasked = np.ma.masked_where(H == 0, H)  # Mask pixels with a value of zero
                # Plot 2D histogram using pcolor
                ax_data = axes.pcolormesh(xedges, yedges, Hmasked, cmap="jet", vmin=1, linewidth=0)
        else:
            raise ValueError(f"没有type: {type}参数对应的画图模式")

        # 生成包络线图层
        axes.plot(
            [-1 / 3, xy_max_edge + 1],
            [-1 / 3, (xy_max_edge + 1) - (0.05 + 0.15 * (xy_max_edge + 1))],
            linestyle="--",
            color="black",
            label="EE envelopes",
            linewidth=0.5,
        )
        axes.plot(
            [-1 / 3, xy_max_edge],
            [-1 / 3, xy_max_edge + (0.05 + 0.15 * xy_max_edge)],
            linestyle="--",
            color="black",
            linewidth=0.6,
        )
        axes.plot([-1 / 3, xy_max_edge + 1], [-1 / 3, xy_max_edge + 1], linestyle="-", color="black", linewidth=0.5)

        # 生成文本图层
        # # 计算指标
        above_percent, below_percent, within_percent = envelope_of_expected_error(x, y)
        rmse = root_mean_square_error(x, y)
        r2 = r2_score(x, y) if len(x) > 1 else 0
        add_text = "\n".join(
            [
                f"$EE$ envelopes: ±(0.05+15%)",
                f"     {format(within_percent, '.1f')}% within $EE$",
                f"     {format(above_percent, '.1f')}% above $EE$",
                f"     {format(below_percent, '.1f')}% below $EE$",
                f"$RMSE$ = {format(rmse, '.3f')}",
                f"$R^2$ = {format(r2, '.3f')}",
                f"$N$ = {len(x)}",
            ]
        )
        if style in ["single_normal", "single_bold"]:
            alpha_text = 0.7
        else:
            alpha_text = 0
        axes.text(
            x=0,
            y=xy_max_edge,
            s=add_text,
            fontsize=ticksize,
            verticalalignment="top",
            color="black",
            bbox=dict(boxstyle="round", facecolor="white", alpha=alpha_text, linewidth=0.2),
        )

        # 生成图例
        lg = axes.legend(
            loc=4,
            alignment="right",
            fontsize=ticksize,
            labelspacing=0.2,
            borderpad=0.2,
            handletextpad=0.2,
            handlelength=1,
        )
        lg.get_frame().set(linewidth=0.2, edgecolor="k", alpha=0.5)

        # 生成colorbar
        cb = fig.colorbar(ax_data, ax=axes, pad=0.02, fraction=0.05)
        if type in ["scatter", "grid"]:
            cb.ax.set_title(label="Num", loc="center", fontdict={"size": ticksize}, pad=0)  # loc参数
            yticks = list(filter(lambda x: x == int(x), cb.ax.get_yticks()))
            cb.ax.set_yticks(yticks)
        else:
            cb.ax.set_title(label="Density", loc="center", fontdict={"size": ticksize}, pad=0)
        cb.ax.tick_params(labelsize=ticksize)

        # 画图
        if save_path is not None:
            # 检查保存路径父文件夹是否存在, 不存在则创建上级目录
            pic_dirpath = os.path.dirname(os.path.abspath(save_path))
            if not os.path.exists(pic_dirpath):
                os.makedirs(pic_dirpath)
            # 保存图片
            fig.savefig(save_path, dpi=dpi)
            plt.close(fig)
        else:
            # 显示图片
            plt.show()



import numpy as np
from mtmaod.utils.pyhdf import PyHDF, SDS

from .template import SatelliteProductReader, SatelliteProductData


class MOD04_Data(SatelliteProductData):
    def __init__(self, dp:SDS, isRaw:bool = False, *args, **kwargs) -> None:
        self.dp = dp
        self.isRaw = isRaw
        pass

    def scale_and_offset(self, data:np.ndarray):
        infos:dict = self.infos()
        scale_factor = MOD04_Data.value_set_decimal(infos.get("scale_factor", 1), decimal=8)
        add_offset = MOD04_Data.value_set_decimal(infos.get("add_offset", 0), decimal=8)
        fill_value = infos.get("_FillValue")
        data = data.astype(np.float64)
        data[data == fill_value] = np.nan
        return data * scale_factor + add_offset

    def infos(self):
        return PyHDF.get_dataset_info_from_dp(self.dp)

    def __getitem__(self, *item):
        data = self.dp.__getitem__(*item)
        return self.scale_and_offset(data) if not self.isRaw else data


class MOD04(SatelliteProductReader):
    Product_File_Time_Format = "[.]A%Y%j[.]%H%M[.]"
    Band_Latitude = "Latitude"
    Band_Longitude = "Longitude"

    @staticmethod
    def open(data_file, *args, **kwargs):
        return PyHDF.open(data_file, *args, **kwargs)

    @staticmethod
    def read(fp, dataset_name, *args, isRaw=False, **kwargs):
        dp = PyHDF.read(fp, dataset_name, *args, **kwargs)
        return MOD04_Data(dp, isRaw=isRaw)
    
    @staticmethod
    def list_datasets(fp, full:bool = False, *args, **kwargs):
        return PyHDF.list_datasets(fp, full=full, *args, **kwargs)



from mtmaod.products.mod09 import MOD09
import psutil, os
import numpy as np
from pprint import pprint as print

path = r"C:\Users\imutu\Desktop\MOD09GA.A2021001.h24v04.061.2021012061136.hdf"

ds = MOD09.open(path)
band_names = MOD09.list_datasets(ds, full=True)
print(band_names)

band = "sur_refl_b02_1"
data = MOD09.read(ds, band)[:]
print(data)
print((data.dtype, data.shape))
# band_aod = "Optical_Depth_Land_And_Ocean"
# data_aod = MOD09.read(ds, band_aod)[:]
# print(data_aod.dtype, data_aod.shape)

# data_lat = MOD09.read(ds, MOD09.Band_Latitude)[:]
# print(data_lat.dtype, data_lat.shape)

# data_lon = MOD09.read(ds, MOD09.Band_Longitude)[:]
# print(data_lon)
# print(data_lon.dtype, data_lon.shape)

# print(u'当前进程的内存使用：%.4f GB' % (psutil.Process(os.getpid()).memory_info().rss / 1024 / 1024 / 1024))